uses windows
/* scalar types  */
typedef char       cl_char;
typedef u8         cl_uchar;
typedef short      cl_short;
typedef u16        cl_ushort;
typedef int        cl_int;
typedef u32        cl_uint;
typedef i64        cl_long;
typedef u64        cl_ulong;

typedef u16        cl_half;
typedef float      cl_float;
typedef double     cl_double;


/*
* Vector types
*
*  Note:   OpenCL requires that all types be naturally aligned.
*          This means that vector types must be naturally aligned.
*          For example, a vector of four floats must be aligned to
*          a 16 byte boundary (calculated as 4 * the natural 4-byte
*          alignment of the float).  The alignment qualifiers here
*          will only function properly if your compiler supports them
*          and if you don't actively work to defeat them.  For example,
*          in order for a cl_float4 to be 16 byte aligned in a struct,
*          the start of the struct must itself be 16-byte aligned.
*
*          Maintaining proper alignment is the user's responsibility.
*/
typedef i8[2]          cl_char2;
typedef i8[4]          cl_char4;
typedef i8[8]          cl_char8;
typedef i8[16]          cl_char16;
typedef u8[2]         cl_uchar2;
typedef u8[4]         cl_uchar4;
typedef u8[8]         cl_uchar8;
typedef u8[16]         cl_uchar16;

typedef i16[2]         cl_short2;
typedef i16[4]         cl_short4;
typedef i16[8]         cl_short8;
typedef i16[16]         cl_short16;
typedef u16[2]        cl_ushort2;
typedef u16[4]        cl_ushort4;
typedef u16[8]        cl_ushort8;
typedef u16[16]        cl_ushort16;

typedef i32[2]         cl_int2;
typedef i32[4]         cl_int4;
typedef i32[8]         cl_int8;
typedef i32[16]         cl_int16;
typedef u32[2]        cl_uint2;
typedef u32[4]        cl_uint4;
typedef u32[8]        cl_uint8;
typedef u32[16]        cl_uint16;

typedef i64[2]         cl_long2;
typedef i64[4]         cl_long4;
typedef i64[8]         cl_long8;
typedef i64[16]         cl_long16;
typedef u64[2]        cl_ulong2;
typedef u64[4]        cl_ulong4;
typedef u64[8]        cl_ulong8;
typedef u64[16]        cl_ulong16;

typedef float[2]           cl_float2;
typedef float[4]           cl_float4;
typedef float[8]           cl_float8;
typedef float[16]           cl_float16;

typedef double[2]          cl_double2;
typedef double[4]          cl_double4;
typedef double[8]          cl_double8;
typedef double[16]          cl_double16;


const CL_CHAR_BIT   =      8
const CL_SCHAR_MAX  =      127
const CL_SCHAR_MIN  =      (-127-1)
const CL_CHAR_MAX   =      CL_SCHAR_MAX
const CL_CHAR_MIN   =      CL_SCHAR_MIN
const CL_UCHAR_MAX  =      255
const CL_SHRT_MAX   =      32767
const CL_SHRT_MIN   =      (-32767-1)
const CL_USHRT_MAX  =      65535
const CL_INT_MAX    =      2147483647
const CL_INT_MIN    =      (-2147483647-1)
const CL_UINT_MAX  =       0xffffffffL
const CL_LONG_MAX  =       ((cl_long) 0x7FFFFFFFFFFFFFFFL)
const CL_LONG_MIN  =       ((cl_long) -0x7FFFFFFFFFFFFFFFL - 1L)
const CL_ULONG_MAX =       ((cl_ulong) 0xFFFFFFFFFFFFFFFFL)

const CL_FLT_DIG         = 6
const CL_FLT_MANT_DIG    = 24
const CL_FLT_MAX_10_EXP  = 38
const CL_FLT_MAX_EXP     = 128
const CL_FLT_MIN_10_EXP  = -37
const CL_FLT_MIN_EXP     = -125
const CL_FLT_RADIX       = 2

// MSVC doesn't understand hex floats
const CL_FLT_MAX      =    3.402823466e38f
const CL_FLT_MIN      =    1.175494351e-38f
const CL_FLT_EPSILON  =    1.192092896e-07f
//#else
//#define CL_FLT_MAX          0x1.fffffep127f
//#define CL_FLT_MIN          0x1.0p-126f
//#define CL_FLT_EPSILON      0x1.0p-23f
//#endif

const CL_DBL_DIG         = 15
const CL_DBL_MANT_DIG    = 53
const CL_DBL_MAX_10_EXP  = 308
const CL_DBL_MAX_EXP     = 1024
const CL_DBL_MIN_10_EXP  = -307
const CL_DBL_MIN_EXP     = -1021
const CL_DBL_RADIX       = 2
// MSVC doesn't understand hex floats
const CL_DBL_MAX         = 1.7976931348623158e308
const CL_DBL_MIN         = 2.2250738585072014e-308
const CL_DBL_EPSILON     = 2.2204460492503131e-016
//#else
//#define CL_DBL_MAX          0x1.fffffffffffffp1023
//#define CL_DBL_MIN          0x1.0p-1022
//#define CL_DBL_EPSILON      0x1.0p-52
//#endif

//  CL.h contents
/******************************************************************************/
struct _cl_platform_id
struct _cl_device_id
struct _cl_context
struct _cl_command_queue
struct _cl_mem
struct _cl_program
struct _cl_kernel
struct _cl_event
struct _cl_sampler

typedef  _cl_platform_id *  cl_platform_id;
typedef  _cl_device_id *  cl_device_id;
typedef  _cl_context *  cl_context;
typedef  _cl_command_queue *  cl_command_queue;
typedef  _cl_mem *  cl_mem;
typedef  _cl_program *  cl_program;
typedef  _cl_kernel *  cl_kernel;
typedef  _cl_event *  cl_event;
typedef  _cl_sampler *  cl_sampler;

typedef cl_uint             cl_bool;                     /* WARNING!  Unlike cl_ types in cl_platform.h, cl_bool is not guaranteed to be the same size as the bool in kernels. */
typedef cl_ulong            cl_bitfield;
typedef cl_bitfield         cl_device_type;
typedef cl_uint             cl_platform_info;
typedef cl_uint             cl_device_info;
typedef cl_bitfield         cl_device_address_info;
typedef cl_bitfield         cl_device_fp_config;
typedef cl_uint             cl_device_mem_cache_type;
typedef cl_uint             cl_device_local_mem_type;
typedef cl_bitfield         cl_device_exec_capabilities;
typedef cl_bitfield         cl_command_queue_properties;

typedef void*			cl_context_properties;
typedef cl_uint             cl_context_info;
typedef cl_uint             cl_command_queue_info;
typedef cl_uint             cl_channel_order;
typedef cl_uint             cl_channel_type;
typedef cl_bitfield         cl_mem_flags;
typedef cl_uint             cl_mem_object_type;
typedef cl_uint             cl_mem_info;
typedef cl_uint             cl_image_info;
typedef cl_uint             cl_addressing_mode;
typedef cl_uint             cl_filter_mode;
typedef cl_uint             cl_sampler_info;
typedef cl_bitfield         cl_map_flags;
typedef cl_uint             cl_program_info;
typedef cl_uint             cl_program_build_info;
typedef cl_int              cl_build_status;
typedef cl_uint             cl_kernel_info;
typedef cl_uint             cl_kernel_work_group_info;
typedef cl_uint             cl_event_info;
typedef cl_uint             cl_command_type;
typedef cl_uint             cl_profiling_info;

struct _cl_image_format
    cl_channel_order        image_channel_order;
    cl_channel_type         image_channel_data_type;
typedef _cl_image_format cl_image_format;



/******************************************************************************/

// Error Codes
const CL_SUCCESS =                                  0
const CL_DEVICE_NOT_FOUND =                         -1
const CL_DEVICE_NOT_AVAILABLE =                     -2
const CL_COMPILER_NOT_AVAILABLE =                   -3
const CL_MEM_OBJECT_ALLOCATION_FAILURE =            -4
const CL_OUT_OF_RESOURCES =                         -5
const CL_OUT_OF_HOST_MEMORY =                       -6
const CL_PROFILING_INFO_NOT_AVAILABLE =             -7
const CL_MEM_COPY_OVERLAP =                         -8
const CL_IMAGE_FORMAT_MISMATCH =                    -9
const CL_IMAGE_FORMAT_NOT_SUPPORTED =               -10
const CL_BUILD_PROGRAM_FAILURE =                    -11
const CL_MAP_FAILURE =                              -12

const CL_INVALID_VALUE =                            -30
const CL_INVALID_DEVICE_TYPE =                      -31
const CL_INVALID_PLATFORM =                         -32
const CL_INVALID_DEVICE =                           -33
const CL_INVALID_CONTEXT =                          -34
const CL_INVALID_QUEUE_PROPERTIES =                 -35
const CL_INVALID_COMMAND_QUEUE =                    -36
const CL_INVALID_HOST_PTR =                         -37
const CL_INVALID_MEM_OBJECT =                       -38
const CL_INVALID_IMAGE_FORMAT_DESCRIPTOR =          -39
const CL_INVALID_IMAGE_SIZE =                       -40
const CL_INVALID_SAMPLER =                          -41
const CL_INVALID_BINARY =                           -42
const CL_INVALID_BUILD_OPTIONS =                    -43
const CL_INVALID_PROGRAM =                          -44
const CL_INVALID_PROGRAM_EXECUTABLE =               -45
const CL_INVALID_KERNEL_NAME =                      -46
const CL_INVALID_KERNEL_DEFINITION =                -47
const CL_INVALID_KERNEL =                           -48
const CL_INVALID_ARG_INDEX =                        -49
const CL_INVALID_ARG_VALUE =                        -50
const CL_INVALID_ARG_SIZE =                         -51
const CL_INVALID_KERNEL_ARGS =                      -52
const CL_INVALID_WORK_DIMENSION =                   -53
const CL_INVALID_WORK_GROUP_SIZE =                  -54
const CL_INVALID_WORK_ITEM_SIZE =                   -55
const CL_INVALID_GLOBAL_OFFSET =                    -56
const CL_INVALID_EVENT_WAIT_LIST =                  -57
const CL_INVALID_EVENT =                            -58
const CL_INVALID_OPERATION =                        -59
const CL_INVALID_GL_OBJECT =                        -60
const CL_INVALID_BUFFER_SIZE =                      -61
const CL_INVALID_MIP_LEVEL =                        -62
const CL_INVALID_GLOBAL_WORK_SIZE =                 -63

// OpenCL Version
const CL_VERSION_1_0 =                              1

// cl_bool
const u32 CL_FALSE =                                    0
const u32 CL_TRUE =                                     1

// cl_platform_info
const CL_PLATFORM_PROFILE =                         0x0900
const CL_PLATFORM_VERSION =                         0x0901
const CL_PLATFORM_NAME =                            0x0902
const CL_PLATFORM_VENDOR =                          0x0903
const CL_PLATFORM_EXTENSIONS =                      0x0904

// cl_device_type - bitfield
const cl_device_type CL_DEVICE_TYPE_DEFAULT =                      (1 << 0)
const cl_device_type CL_DEVICE_TYPE_CPU =                          (1 << 1)
const cl_device_type CL_DEVICE_TYPE_GPU =                          (1 << 2)
const cl_device_type CL_DEVICE_TYPE_ACCELERATOR =                  (1 << 3)
const cl_device_type CL_DEVICE_TYPE_ALL =                          0xFFFFFFFF

// cl_device_info
const cl_device_info CL_DEVICE_TYPE =                              0x1000
const cl_device_info CL_DEVICE_VENDOR_ID =                         0x1001
const cl_device_info CL_DEVICE_MAX_COMPUTE_UNITS =                 0x1002
const cl_device_info CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS =          0x1003
const cl_device_info CL_DEVICE_MAX_WORK_GROUP_SIZE =               0x1004
const cl_device_info CL_DEVICE_MAX_WORK_ITEM_SIZES =               0x1005
const cl_device_info CL_DEVICE_PREFERRED_VECTOR_WIDTH_CHAR =       0x1006
const cl_device_info CL_DEVICE_PREFERRED_VECTOR_WIDTH_SHORT =      0x1007
const cl_device_info CL_DEVICE_PREFERRED_VECTOR_WIDTH_INT =        0x1008
const cl_device_info CL_DEVICE_PREFERRED_VECTOR_WIDTH_LONG =       0x1009
const cl_device_info CL_DEVICE_PREFERRED_VECTOR_WIDTH_FLOAT =      0x100A
const cl_device_info CL_DEVICE_PREFERRED_VECTOR_WIDTH_DOUBLE =     0x100B
const cl_device_info CL_DEVICE_MAX_CLOCK_FREQUENCY =               0x100C
const cl_device_info CL_DEVICE_ADDRESS_BITS =                      0x100D
const cl_device_info CL_DEVICE_MAX_READ_IMAGE_ARGS =               0x100E
const cl_device_info CL_DEVICE_MAX_WRITE_IMAGE_ARGS =              0x100F
const cl_device_info CL_DEVICE_MAX_MEM_ALLOC_SIZE =                0x1010
const cl_device_info CL_DEVICE_IMAGE2D_MAX_WIDTH =                 0x1011
const cl_device_info CL_DEVICE_IMAGE2D_MAX_HEIGHT =                0x1012
const cl_device_info CL_DEVICE_IMAGE3D_MAX_WIDTH =                 0x1013
const cl_device_info CL_DEVICE_IMAGE3D_MAX_HEIGHT =                0x1014
const cl_device_info CL_DEVICE_IMAGE3D_MAX_DEPTH =                 0x1015
const cl_device_info CL_DEVICE_IMAGE_SUPPORT =                     0x1016
const cl_device_info CL_DEVICE_MAX_PARAMETER_SIZE =                0x1017
const cl_device_info CL_DEVICE_MAX_SAMPLERS =                      0x1018
const cl_device_info CL_DEVICE_MEM_BASE_ADDR_ALIGN =               0x1019
const cl_device_info CL_DEVICE_MIN_DATA_TYPE_ALIGN_SIZE =          0x101A
const cl_device_info CL_DEVICE_SINGLE_FP_CONFIG =                  0x101B
const cl_device_info CL_DEVICE_GLOBAL_MEM_CACHE_TYPE =             0x101C
const cl_device_info CL_DEVICE_GLOBAL_MEM_CACHELINE_SIZE =         0x101D
const cl_device_info CL_DEVICE_GLOBAL_MEM_CACHE_SIZE =             0x101E
const cl_device_info CL_DEVICE_GLOBAL_MEM_SIZE =                   0x101F
const cl_device_info CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE =          0x1020
const cl_device_info CL_DEVICE_MAX_CONSTANT_ARGS =                 0x1021
const cl_device_info CL_DEVICE_LOCAL_MEM_TYPE =                    0x1022
const cl_device_info CL_DEVICE_LOCAL_MEM_SIZE =                    0x1023
const cl_device_info CL_DEVICE_ERROR_CORRECTION_SUPPORT =          0x1024
const cl_device_info CL_DEVICE_PROFILING_TIMER_RESOLUTION =        0x1025
const cl_device_info CL_DEVICE_ENDIAN_LITTLE =                     0x1026
const cl_device_info CL_DEVICE_AVAILABLE =                         0x1027
const cl_device_info CL_DEVICE_COMPILER_AVAILABLE =                0x1028
const cl_device_info CL_DEVICE_EXECUTION_CAPABILITIES =            0x1029
const cl_device_info CL_DEVICE_QUEUE_PROPERTIES =                  0x102A
const cl_device_info CL_DEVICE_NAME =                              0x102B
const cl_device_info CL_DEVICE_VENDOR =                            0x102C
const cl_device_info CL_DRIVER_VERSION =                           0x102D
const cl_device_info CL_DEVICE_PROFILE =                           0x102E
const cl_device_info CL_DEVICE_VERSION =                           0x102F
const cl_device_info CL_DEVICE_EXTENSIONS =                        0x1030
const cl_device_info CL_DEVICE_PLATFORM =                          0x1031

// cl_device_fp_config - bitfield
const cl_device_fp_config CL_FP_DENORM =                                (1 << 0)
const cl_device_fp_config CL_FP_INF_NAN =                               (1 << 1)
const cl_device_fp_config CL_FP_ROUND_TO_NEAREST =                      (1 << 2)
const cl_device_fp_config CL_FP_ROUND_TO_ZERO =                         (1 << 3)
const cl_device_fp_config CL_FP_ROUND_TO_INF =                          (1 << 4)
const cl_device_fp_config CL_FP_FMA =                                   (1 << 5)

// cl_device_mem_cache_type
const cl_device_mem_cache_type CL_NONE =                                     0x0
const cl_device_mem_cache_type CL_READ_ONLY_CACHE =                          0x1
const cl_device_mem_cache_type CL_READ_WRITE_CACHE =                         0x2

// cl_device_local_mem_type
const cl_device_local_mem_type CL_LOCAL =                                    0x1
const cl_device_local_mem_type CL_GLOBAL =                                   0x2

// cl_device_exec_capabilities - bitfield
const cl_device_exec_capabilities CL_EXEC_KERNEL =                              (1 << 0)
const cl_device_exec_capabilities CL_EXEC_NATIVE_KERNEL =                       (1 << 1)

// cl_command_queue_properties - bitfield
const cl_command_queue_properties CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE =      (1 << 0)
const cl_command_queue_properties CL_QUEUE_PROFILING_ENABLE =                   (1 << 1)

// cl_context_info
const cl_context_info CL_CONTEXT_REFERENCE_COUNT =                  0x1080
const cl_context_info CL_CONTEXT_DEVICES =                          0x1081
const cl_context_info CL_CONTEXT_PROPERTIES =                       0x1082

// cl_context_properties
const cl_context_properties CL_CONTEXT_PLATFORM =                         0x1084

// cl_command_queue_info
const cl_command_queue_info CL_QUEUE_CONTEXT =                            0x1090
const cl_command_queue_info CL_QUEUE_DEVICE =                             0x1091
const cl_command_queue_info CL_QUEUE_REFERENCE_COUNT =                    0x1092
const cl_command_queue_info CL_QUEUE_PROPERTIES =                         0x1093

// cl_mem_flags - bitfield
const cl_mem_flags CL_MEM_READ_WRITE =                           (1 << 0)
const cl_mem_flags CL_MEM_WRITE_ONLY =                           (1 << 1)
const cl_mem_flags CL_MEM_READ_ONLY =                            (1 << 2)
const cl_mem_flags CL_MEM_USE_HOST_PTR =                         (1 << 3)
const cl_mem_flags CL_MEM_ALLOC_HOST_PTR =                       (1 << 4)
const cl_mem_flags CL_MEM_COPY_HOST_PTR =                        (1 << 5)

// cl_channel_order
const cl_channel_order CL_R =                                        0x10B0
const cl_channel_order CL_A =                                        0x10B1
const cl_channel_order CL_RG =                                       0x10B2
const cl_channel_order CL_RA =                                       0x10B3
const cl_channel_order CL_RGB =                                      0x10B4
const cl_channel_order CL_RGBA =                                     0x10B5
const cl_channel_order CL_BGRA =                                     0x10B6
const cl_channel_order CL_ARGB =                                     0x10B7
const cl_channel_order CL_INTENSITY =                                0x10B8
const cl_channel_order CL_LUMINANCE =                                0x10B9

// cl_channel_type
const cl_channel_type CL_SNORM_INT8 =                               0x10D0
const cl_channel_type CL_SNORM_INT16 =                              0x10D1
const cl_channel_type CL_UNORM_INT8 =                               0x10D2
const cl_channel_type CL_UNORM_INT16 =                              0x10D3
const cl_channel_type CL_UNORM_SHORT_565 =                          0x10D4
const cl_channel_type CL_UNORM_SHORT_555 =                          0x10D5
const cl_channel_type CL_UNORM_INT_101010 =                         0x10D6
const cl_channel_type CL_SIGNED_INT8 =                              0x10D7
const cl_channel_type CL_SIGNED_INT16 =                             0x10D8
const cl_channel_type CL_SIGNED_INT32 =                             0x10D9
const cl_channel_type CL_UNSIGNED_INT8 =                            0x10DA
const cl_channel_type CL_UNSIGNED_INT16 =                           0x10DB
const cl_channel_type CL_UNSIGNED_INT32 =                           0x10DC
const cl_channel_type CL_HALF_FLOAT =                               0x10DD
const cl_channel_type CL_FLOAT =                                    0x10DE

// cl_mem_object_type
const cl_mem_object_type CL_MEM_OBJECT_BUFFER =                        0x10F0
const cl_mem_object_type CL_MEM_OBJECT_IMAGE2D =                       0x10F1
const cl_mem_object_type CL_MEM_OBJECT_IMAGE3D =                       0x10F2

// cl_mem_info
const cl_mem_info CL_MEM_TYPE =                                 0x1100
const cl_mem_info CL_MEM_FLAGS =                                0x1101
const cl_mem_info CL_MEM_SIZE =                                 0x1102
const cl_mem_info CL_MEM_HOST_PTR =                             0x1103
const cl_mem_info CL_MEM_MAP_COUNT =                            0x1104
const cl_mem_info CL_MEM_REFERENCE_COUNT =                      0x1105
const cl_mem_info CL_MEM_CONTEXT =                              0x1106

// cl_image_info
const cl_image_info CL_IMAGE_FORMAT =                             0x1110
const cl_image_info CL_IMAGE_ELEMENT_SIZE =                       0x1111
const cl_image_info CL_IMAGE_ROW_PITCH =                          0x1112
const cl_image_info CL_IMAGE_SLICE_PITCH =                        0x1113
const cl_image_info CL_IMAGE_WIDTH =                              0x1114
const cl_image_info CL_IMAGE_HEIGHT =                             0x1115
const cl_image_info CL_IMAGE_DEPTH =                              0x1116

// cl_addressing_mode
const cl_addressing_mode CL_ADDRESS_NONE =                             0x1130
const cl_addressing_mode CL_ADDRESS_CLAMP_TO_EDGE =                    0x1131
const cl_addressing_mode CL_ADDRESS_CLAMP =                            0x1132
const cl_addressing_mode CL_ADDRESS_REPEAT =                           0x1133

// cl_filter_mode
const cl_filter_mode CL_FILTER_NEAREST =                           0x1140
const cl_filter_mode CL_FILTER_LINEAR =                            0x1141

// cl_sampler_info
const cl_sampler_info CL_SAMPLER_REFERENCE_COUNT =                  0x1150
const cl_sampler_info CL_SAMPLER_CONTEXT =                          0x1151
const cl_sampler_info CL_SAMPLER_NORMALIZED_COORDS =                0x1152
const cl_sampler_info CL_SAMPLER_ADDRESSING_MODE =                  0x1153
const cl_sampler_info CL_SAMPLER_FILTER_MODE =                      0x1154

// cl_map_flags - bitfield
const cl_map_flags CL_MAP_READ =                                 (1 << 0)
const cl_map_flags CL_MAP_WRITE =                                (1 << 1)

// cl_program_info
const cl_program_info CL_PROGRAM_REFERENCE_COUNT =                  0x1160
const cl_program_info CL_PROGRAM_CONTEXT =                          0x1161
const cl_program_info CL_PROGRAM_NUM_DEVICES =                      0x1162
const cl_program_info CL_PROGRAM_DEVICES =                          0x1163
const cl_program_info CL_PROGRAM_SOURCE =                           0x1164
const cl_program_info CL_PROGRAM_BINARY_SIZES =                     0x1165
const cl_program_info CL_PROGRAM_BINARIES =                         0x1166

// cl_program_build_info
const cl_program_build_info CL_PROGRAM_BUILD_STATUS =                     0x1181
const cl_program_build_info CL_PROGRAM_BUILD_OPTIONS =                    0x1182
const cl_program_build_info CL_PROGRAM_BUILD_LOG =                        0x1183

// cl_build_status
const cl_build_status CL_BUILD_SUCCESS =                            0
const cl_build_status CL_BUILD_NONE =                               -1
const cl_build_status CL_BUILD_ERROR =                              -2
const cl_build_status CL_BUILD_IN_PROGRESS =                        -3

// cl_kernel_info
const cl_kernel_info CL_KERNEL_FUNCTION_NAME =                     0x1190
const cl_kernel_info CL_KERNEL_NUM_ARGS =                          0x1191
const cl_kernel_info CL_KERNEL_REFERENCE_COUNT =                   0x1192
const cl_kernel_info CL_KERNEL_CONTEXT =                           0x1193
const cl_kernel_info CL_KERNEL_PROGRAM =                           0x1194

// cl_kernel_work_group_info
const cl_kernel_work_group_info CL_KERNEL_WORK_GROUP_SIZE =                   0x11B0
const cl_kernel_work_group_info CL_KERNEL_COMPILE_WORK_GROUP_SIZE =           0x11B1
const cl_kernel_work_group_info CL_KERNEL_LOCAL_MEM_SIZE =                    0x11B2

// cl_event_info
const cl_event_info CL_EVENT_COMMAND_QUEUE =                      0x11D0
const cl_event_info CL_EVENT_COMMAND_TYPE =                       0x11D1
const cl_event_info CL_EVENT_REFERENCE_COUNT =                    0x11D2
const cl_event_info CL_EVENT_COMMAND_EXECUTION_STATUS =           0x11D3

// cl_command_type
const cl_command_type CL_COMMAND_NDRANGE_KERNEL =                   0x11F0
const cl_command_type CL_COMMAND_TASK =                             0x11F1
const cl_command_type CL_COMMAND_NATIVE_KERNEL =                    0x11F2
const cl_command_type CL_COMMAND_READ_BUFFER =                      0x11F3
const cl_command_type CL_COMMAND_WRITE_BUFFER =                     0x11F4
const cl_command_type CL_COMMAND_COPY_BUFFER =                      0x11F5
const cl_command_type CL_COMMAND_READ_IMAGE =                       0x11F6
const cl_command_type CL_COMMAND_WRITE_IMAGE =                      0x11F7
const cl_command_type CL_COMMAND_COPY_IMAGE =                       0x11F8
const cl_command_type CL_COMMAND_COPY_IMAGE_TO_BUFFER =             0x11F9
const cl_command_type CL_COMMAND_COPY_BUFFER_TO_IMAGE =             0x11FA
const cl_command_type CL_COMMAND_MAP_BUFFER =                       0x11FB
const cl_command_type CL_COMMAND_MAP_IMAGE =                        0x11FC
const cl_command_type CL_COMMAND_UNMAP_MEM_OBJECT =                 0x11FD
const cl_command_type CL_COMMAND_MARKER =                           0x11FE
const cl_command_type CL_COMMAND_ACQUIRE_GL_OBJECTS =               0x11FF
const cl_command_type CL_COMMAND_RELEASE_GL_OBJECTS =               0x1200

// command execution status
const CL_COMPLETE =                                 0x0
const CL_RUNNING =                                  0x1
const CL_SUBMITTED =                                0x2
const CL_QUEUED =                                   0x3

// cl_profiling_info
const cl_profiling_info CL_PROFILING_COMMAND_QUEUED =                 0x1280
const cl_profiling_info CL_PROFILING_COMMAND_SUBMIT =                 0x1281
const cl_profiling_info CL_PROFILING_COMMAND_START =                  0x1282
const cl_profiling_info CL_PROFILING_COMMAND_END =                    0x1283

/********************************************************************************************************/

/********************************************************************************************************/

//  Function signature typedef's

// Platform API
typedef cl_int()stdcall(cl_uint  /* num_entries */,   cl_platform_id * /* platforms */,   cl_uint *  /* num_platforms */) PFNCLGETPLATFORMIDS
typedef cl_int()stdcall(cl_platform_id   /* platform */,  cl_platform_info /* param_name */,  size_t   /* param_value_size */,  void *   /* param_value */,  size_t * /* param_value_size_ret */) PFNCLGETPLATFORMINFO
// Device APIs
typedef cl_int()stdcall(cl_platform_id   /* platform */, cl_device_type   /* device_type */, cl_uint  /* num_entries */, cl_device_id *   /* devices */, cl_uint *  /* num_devices */) PFNCLGETDEVICEIDS
typedef cl_int()stdcall(cl_device_id    /* device */,  cl_device_info  /* param_name */,  size_t  /* param_value_size */,  void *  /* param_value */,  size_t *  /* param_value_size_ret */) PFNCLGETDEVICEINFO
// Context APIs
typedef void()stdcall (char *, void *, size_t, void *) pfn_notify
typedef cl_context()stdcall(cl_context_properties * /* properties */,  cl_uint /* num_devices */,  cl_device_id *  /* devices */,  pfn_notify /* pfn_notify */,  void *  /* user_data */,  cl_int *  /* errcode_ret */) PFNCLCREATECONTEXT
typedef cl_context()stdcall(cl_context_properties * /* properties */,  cl_device_type  /* device_type */,  pfn_notify /* pfn_notify */,  void *  /* user_data */,  cl_int *  /* errcode_ret */) PFNCLCREATECONTEXTFROMTYPE
typedef cl_int()stdcall(cl_context /* context */) PFNCLRETAINCONTEXT
typedef cl_int()stdcall(cl_context /* context */) PFNCLRELEASECONTEXT
typedef cl_int()stdcall(cl_context /* context */,   cl_context_info    /* param_name */,   size_t /* param_value_size */,   void * /* param_value */,   size_t *   /* param_value_size_ret */) PFNCLGETCONTEXTINFO
// Command Queue APIs
typedef cl_command_queue()stdcall(cl_context /* context */, cl_device_id   /* device */, cl_command_queue_properties    /* properties */, cl_int * /* errcode_ret */) PFNCLCREATECOMMANDQUEUE
typedef cl_int()stdcall(cl_command_queue /* command_queue */) PFNCLRETAINCOMMANDQUEUE
typedef cl_int()stdcall(cl_command_queue /* command_queue */) PFNCLRELEASECOMMANDQUEUE
typedef cl_int()stdcall(cl_command_queue  /* command_queue */,  cl_command_queue_info /* param_name */,  size_t  /* param_value_size */,  void *  /* param_value */,  size_t *  /* param_value_size_ret */) PFNCLGETCOMMANDQUEUEINFO
typedef cl_int()stdcall(cl_command_queue  /* command_queue */,    cl_command_queue_properties   /* properties */,    cl_bool  /* enable */,    cl_command_queue_properties * /* old_properties */) PFNCLSETCOMMANDQUEUEPROPERTY
// Memory Object APIs
typedef cl_mem()stdcall(cl_context   /* context */, cl_mem_flags /* flags */, size_t /* size */, void * /* host_ptr */, cl_int * /* errcode_ret */) PFNCLCREATEBUFFER
typedef cl_mem()stdcall(cl_context  /* context */,  cl_mem_flags    /* flags */,  cl_image_format * /* image_format */,  size_t  /* image_width */,  size_t  /* image_height */,  size_t  /* image_row_pitch */,  void *  /* host_ptr */,  cl_int *  /* errcode_ret */) PFNCLCREATEIMAGE2D
typedef cl_mem()stdcall(cl_context  /* context */,  cl_mem_flags    /* flags */,  cl_image_format * /* image_format */,  size_t  /* image_width */,  size_t  /* image_height */,  size_t  /* image_depth */,  size_t  /* image_row_pitch */,  size_t  /* image_slice_pitch */,  void *  /* host_ptr */,  cl_int *  /* errcode_ret */) PFNCLCREATEIMAGE3D
typedef cl_int()stdcall(cl_mem /* memobj */) PFNCLRETAINMEMOBJECT
typedef cl_int()stdcall(cl_mem /* memobj */) PFNCLRELEASEMEMOBJECT
typedef cl_int()stdcall(cl_context   /* context */,   cl_mem_flags /* flags */,   cl_mem_object_type   /* image_type */,   cl_uint  /* num_entries */,   cl_image_format *    /* image_formats */,   cl_uint *    /* num_image_formats */) PFNCLGETSUPPORTEDIMAGEFORMATS
typedef cl_int()stdcall(cl_mem   /* memobj */,   cl_mem_info  /* param_name */,   size_t   /* param_value_size */,   void *   /* param_value */,   size_t * /* param_value_size_ret */) PFNCLGETMEMOBJECTINFO
typedef cl_int()stdcall(cl_mem   /* image */, cl_image_info    /* param_name */, size_t   /* param_value_size */, void *   /* param_value */, size_t * /* param_value_size_ret */) PFNCLGETIMAGEINFO
// Sampler APIs
typedef cl_sampler()stdcall(cl_context  /* context */,  cl_bool /* normalized_coords */,  cl_addressing_mode  /* addressing_mode */,  cl_filter_mode  /* filter_mode */,  cl_int *    /* errcode_ret */) PFNCLCREATESAMPLER
typedef cl_int()stdcall(cl_sampler /* sampler */) PFNCLRETAINSAMPLER
typedef cl_int()stdcall(cl_sampler /* sampler */) PFNCLRELEASESAMPLER
typedef cl_int()stdcall(cl_sampler /* sampler */,   cl_sampler_info    /* param_name */,   size_t /* param_value_size */,   void * /* param_value */,   size_t *   /* param_value_size_ret */) PFNCLGETSAMPLERINFO
// Program Object APIs
typedef cl_program()stdcall(cl_context  /* context */,    cl_uint   /* count */,    char ** /* strings */,    size_t *    /* lengths */,    cl_int *  /* errcode_ret */) PFNCLCREATEPROGRAMWITHSOURCE
typedef cl_program()stdcall(cl_context /* context */,    cl_uint  /* num_devices */,    cl_device_id *   /* device_list */,    size_t *   /* lengths */,    unsigned char ** /* binaries */,    cl_int * /* binary_status */,    cl_int * /* errcode_ret */) PFNCLCREATEPROGRAMWITHBINARY
typedef cl_int()stdcall(cl_program /* program */) PFNCLRETAINPROGRAM
typedef cl_int()stdcall(cl_program /* program */) PFNCLRELEASEPROGRAM
typedef void()stdcall(cl_program /* program */, void * /* user_data */) pfn_notify_build
typedef cl_int()stdcall(cl_program   /* program */, cl_uint  /* num_devices */, cl_device_id * /* device_list */, char * /* options */, pfn_notify_build, void * /* user_data */) PFNCLBUILDPROGRAM
typedef cl_int()stdcall(void) PFNCLUNLOADCOMPILER
typedef cl_int()stdcall(cl_program /* program */,   cl_program_info    /* param_name */,   size_t /* param_value_size */,   void * /* param_value */,   size_t *   /* param_value_size_ret */) PFNCLGETPROGRAMINFO
typedef cl_int()stdcall(cl_program    /* program */,  cl_device_id  /* device */,  cl_program_build_info /* param_name */,  size_t  /* param_value_size */,  void *  /* param_value */,  size_t *  /* param_value_size_ret */) PFNCLGETPROGRAMBUILDINFO
// Kernel Object APIs
typedef cl_kernel()stdcall(cl_program  /* program */, char *    /* kernel_name */, cl_int *  /* errcode_ret */) PFNCLCREATEKERNEL
typedef cl_int()stdcall(cl_program /* program */,   cl_uint  /* num_kernels */,   cl_kernel *    /* kernels */,   cl_uint *  /* num_kernels_ret */) PFNCLCREATEKERNELSINPROGRAM
typedef cl_int()stdcall(cl_kernel    /* kernel */) PFNCLRETAINKERNEL
typedef cl_int()stdcall(cl_kernel   /* kernel */) PFNCLRELEASEKERNEL
typedef cl_int()stdcall(cl_kernel    /* kernel */, cl_uint  /* arg_index */, size_t /* arg_size */, void * /* arg_value */) PFNCLSETKERNELARG
typedef cl_int()stdcall(cl_kernel /* kernel */,  cl_kernel_info  /* param_name */,  size_t  /* param_value_size */,  void *  /* param_value */,  size_t *  /* param_value_size_ret */) PFNCLGETKERNELINFO
typedef cl_int()stdcall(cl_kernel  /* kernel */,   cl_device_id /* device */,   cl_kernel_work_group_info  /* param_name */,   size_t /* param_value_size */,   void * /* param_value */,   size_t *   /* param_value_size_ret */) PFNCLGETKERNELWORKGROUPINFO
// Event Object APIs
typedef cl_int()stdcall(cl_uint /* num_events */,  cl_event *    /* event_list */) PFNCLWAITFOREVENTS
typedef cl_int()stdcall(cl_event /* event */, cl_event_info    /* param_name */, size_t   /* param_value_size */, void *   /* param_value */, size_t * /* param_value_size_ret */) PFNCLGETEVENTINFO
typedef cl_int()stdcall(cl_event /* event */) PFNCLRETAINEVENT
typedef cl_int()stdcall(cl_event /* event */) PFNCLRELEASEEVENT
// Profiling APIs
typedef cl_int()stdcall(cl_event    /* event */,  cl_profiling_info   /* param_name */,  size_t  /* param_value_size */,  void *  /* param_value */,  size_t *    /* param_value_size_ret */) PFNCLGETEVENTPROFILINGINFO
// Flush and Finish APIs
typedef cl_int()stdcall(cl_command_queue /* command_queue */) PFNCLFLUSH
typedef cl_int()stdcall(cl_command_queue /* command_queue */) PFNCLFINISH
// Enqueued Commands APIs
typedef cl_int()stdcall(cl_command_queue    /* command_queue */,    cl_mem  /* buffer */,    cl_bool /* blocking_read */,    size_t  /* offset */,    size_t  /* cb */,    void *  /* ptr */,    cl_uint /* num_events_in_wait_list */,    cl_event *    /* event_wait_list */,    cl_event *  /* event */) PFNCLENQUEUEREADBUFFER
typedef cl_int()stdcall(cl_command_queue   /* command_queue */, cl_mem /* buffer */, cl_bool    /* blocking_write */, size_t /* offset */, size_t /* cb */, void * /* ptr */, cl_uint    /* num_events_in_wait_list */, cl_event *   /* event_wait_list */, cl_event * /* event */) PFNCLENQUEUEWRITEBUFFER
typedef cl_int()stdcall(cl_command_queue    /* command_queue */,    cl_mem  /* src_buffer */,    cl_mem  /* dst_buffer */,    size_t  /* src_offset */,    size_t  /* dst_offset */,    size_t  /* cb */,    cl_uint /* num_events_in_wait_list */,    cl_event *    /* event_wait_list */,    cl_event *  /* event */) PFNCLENQUEUECOPYBUFFER
typedef cl_int()stdcall(cl_command_queue /* command_queue */,   cl_mem /* image */,   cl_bool  /* blocking_read */,   size_t * /* origin[3] */,   size_t * /* region[3] */,   size_t /* row_pitch */,   size_t /* slice_pitch */,   void * /* ptr */,   cl_uint  /* num_events_in_wait_list */,   cl_event * /* event_wait_list */,   cl_event *   /* event */) PFNCLENQUEUEREADIMAGE
typedef cl_int()stdcall(cl_command_queue    /* command_queue */,    cl_mem  /* image */,    cl_bool /* blocking_write */,    size_t *  /* origin[3] */,    size_t *  /* region[3] */,    size_t  /* input_row_pitch */,    size_t  /* input_slice_pitch */,    void *  /* ptr */,    cl_uint /* num_events_in_wait_list */,    cl_event *    /* event_wait_list */,    cl_event *  /* event */) PFNCLENQUEUEWRITEIMAGE
typedef cl_int()stdcall(cl_command_queue /* command_queue */,   cl_mem /* src_image */,   cl_mem /* dst_image */,   size_t * /* src_origin[3] */,   size_t * /* dst_origin[3] */,   size_t * /* region[3] */,   cl_uint  /* num_events_in_wait_list */,   cl_event * /* event_wait_list */,   cl_event *   /* event */) PFNCLENQUEUECOPYIMAGE
typedef cl_int()stdcall(cl_command_queue /* command_queue */,   cl_mem   /* src_image */,   cl_mem   /* dst_buffer */,   size_t *   /* src_origin[3] */,   size_t *   /* region[3] */,   size_t   /* dst_offset */,   cl_uint  /* num_events_in_wait_list */,   cl_event * /* event_wait_list */,   cl_event * /* event */) PFNCLENQUEUECOPYIMAGETOBUFFER
typedef cl_int()stdcall(cl_command_queue /* command_queue */,   cl_mem   /* src_buffer */,   cl_mem   /* dst_image */,   size_t   /* src_offset */,   size_t *   /* dst_origin[3] */,   size_t *   /* region[3] */,   cl_uint  /* num_events_in_wait_list */,   cl_event * /* event_wait_list */,   cl_event * /* event */) PFNCLENQUEUECOPYBUFFERTOIMAGE
typedef void *()(cl_command_queue /* command_queue */,   cl_mem   /* buffer */,   cl_bool  /* blocking_map */,   cl_map_flags /* map_flags */,   size_t   /* offset */,   size_t   /* cb */,   cl_uint  /* num_events_in_wait_list */,   cl_event * /* event_wait_list */,   cl_event * /* event */,   cl_int * /* errcode_ret */) PFNCLENQUEUEMAPBUFFER
typedef void *()(cl_command_queue  /* command_queue */,  cl_mem    /* image */,  cl_bool   /* blocking_map */,  cl_map_flags  /* map_flags */,  size_t *    /* origin[3] */,  size_t *    /* region[3] */,  size_t *  /* image_row_pitch */,  size_t *  /* image_slice_pitch */,  cl_uint   /* num_events_in_wait_list */,  cl_event *  /* event_wait_list */,  cl_event *  /* event */,  cl_int *  /* errcode_ret */)PFNCLENQUEUEMAPIMAGE
typedef cl_int()stdcall(cl_command_queue /* command_queue */,  cl_mem   /* memobj */,  void *   /* mapped_ptr */,  cl_uint  /* num_events_in_wait_list */,  cl_event *  /* event_wait_list */,  cl_event *  /* event */) PFNCLENQUEUEUNMAPMEMOBJECT
typedef cl_int()stdcall(cl_command_queue /* command_queue */, cl_kernel  /* kernel */, cl_uint  /* work_dim */, size_t *   /* global_work_offset */, size_t *   /* global_work_size */, size_t *   /* local_work_size */, cl_uint  /* num_events_in_wait_list */, cl_event * /* event_wait_list */, cl_event * /* event */) PFNCLENQUEUENDRANGEKERNEL
typedef cl_int()stdcall(cl_command_queue  /* command_queue */,  cl_kernel /* kernel */,  cl_uint   /* num_events_in_wait_list */,  cl_event *  /* event_wait_list */,  cl_event *  /* event */) PFNCLENQUEUETASK
typedef void ()(void *) user_func
typedef cl_int()stdcall(cl_command_queue  /* command_queue */,  user_func,  void *    /* args */,  size_t    /* cb_args */,  cl_uint   /* num_mem_objects */,  cl_mem *    /* mem_list */,  void ** /* args_mem_loc */,  cl_uint   /* num_events_in_wait_list */,  cl_event *  /* event_wait_list */,  cl_event *  /* event */)PFNCLENQUEUENATIVEKERNEL
typedef cl_int()stdcall(cl_command_queue    /* command_queue */,  cl_event *  /* event */) PFNCLENQUEUEMARKER
typedef cl_int()stdcall(cl_command_queue /* command_queue */, cl_uint  /* num_events */, cl_event * /* event_list */) PFNCLENQUEUEWAITFOREVENTS
typedef cl_int()stdcall(cl_command_queue /* command_queue */) PFNCLENQUEUEBARRIER
typedef void *()(char * /* func_name */) PFNCLGETEXTENSIONFUNCTIONADDRESS

HMODULE clDllModule = NULL;

//  Variables holding function entry points
PFNCLGETPLATFORMIDS                 clGetPlatformIDs                = NULL;
PFNCLGETPLATFORMINFO                clGetPlatformInfo               = NULL;
PFNCLGETDEVICEIDS                   clGetDeviceIDs                  = NULL;
PFNCLGETDEVICEINFO                  clGetDeviceInfo                 = NULL;
PFNCLCREATECONTEXT                  clCreateContext                 = NULL;
PFNCLCREATECONTEXTFROMTYPE          clCreateContextFromType         = NULL;
PFNCLRETAINCONTEXT                  clRetainContext                 = NULL;
PFNCLRELEASECONTEXT                 clReleaseContext                = NULL;
PFNCLGETCONTEXTINFO                 clGetContextInfo                = NULL;
PFNCLCREATECOMMANDQUEUE             clCreateCommandQueue            = NULL;
PFNCLRETAINCOMMANDQUEUE             clRetainCommandQueue            = NULL;
PFNCLRELEASECOMMANDQUEUE            clReleaseCommandQueue           = NULL;
PFNCLGETCOMMANDQUEUEINFO            clGetCommandQueueInfo           = NULL;
PFNCLSETCOMMANDQUEUEPROPERTY        clSetCommandQueueProperty       = NULL;
PFNCLCREATEBUFFER                   clCreateBuffer                  = NULL;
PFNCLCREATEIMAGE2D                  clCreateImage2D                 = NULL;
PFNCLCREATEIMAGE3D                  clCreateImage3D                 = NULL;
PFNCLRETAINMEMOBJECT                clRetainMemObject               = NULL;
PFNCLRELEASEMEMOBJECT               clReleaseMemObject              = NULL;
PFNCLGETSUPPORTEDIMAGEFORMATS       clGetSupportedImageFormats      = NULL;
PFNCLGETMEMOBJECTINFO               clGetMemObjectInfo              = NULL;
PFNCLGETIMAGEINFO                   clGetImageInfo                  = NULL;
PFNCLCREATESAMPLER                  clCreateSampler                 = NULL;
PFNCLRETAINSAMPLER                  clRetainSampler                 = NULL;
PFNCLRELEASESAMPLER                 clReleaseSampler                = NULL;
PFNCLGETSAMPLERINFO                 clGetSamplerInfo                = NULL;
PFNCLCREATEPROGRAMWITHSOURCE        clCreateProgramWithSource       = NULL;
PFNCLCREATEPROGRAMWITHBINARY        clCreateProgramWithBinary       = NULL;
PFNCLRETAINPROGRAM                  clRetainProgram                 = NULL;
PFNCLRELEASEPROGRAM                 clReleaseProgram                = NULL;
PFNCLBUILDPROGRAM                   clBuildProgram                  = NULL;
PFNCLUNLOADCOMPILER                 clUnloadCompiler                = NULL;
PFNCLGETPROGRAMINFO                 clGetProgramInfo                = NULL;
PFNCLGETPROGRAMBUILDINFO            clGetProgramBuildInfo           = NULL;
PFNCLCREATEKERNEL                   clCreateKernel                  = NULL;
PFNCLCREATEKERNELSINPROGRAM         clCreateKernelsInProgram        = NULL;
PFNCLRETAINKERNEL                   clRetainKernel                  = NULL;
PFNCLRELEASEKERNEL                  clReleaseKernel                 = NULL;
PFNCLSETKERNELARG                   clSetKernelArg                  = NULL;
PFNCLGETKERNELINFO                  clGetKernelInfo                 = NULL;
PFNCLGETKERNELWORKGROUPINFO         clGetKernelWorkGroupInfo        = NULL;
PFNCLWAITFOREVENTS                  clWaitForEvents                 = NULL;
PFNCLGETEVENTINFO                   clGetEventInfo                  = NULL;
PFNCLRETAINEVENT                    clRetainEvent                   = NULL;
PFNCLRELEASEEVENT                   clReleaseEvent                  = NULL;
PFNCLGETEVENTPROFILINGINFO          clGetEventProfilingInfo         = NULL;
PFNCLFLUSH                          clFlush                         = NULL;
PFNCLFINISH                         clFinish                        = NULL;
PFNCLENQUEUEREADBUFFER              clEnqueueReadBuffer             = NULL;
PFNCLENQUEUEWRITEBUFFER             clEnqueueWriteBuffer            = NULL;
PFNCLENQUEUECOPYBUFFER              clEnqueueCopyBuffer             = NULL;
PFNCLENQUEUEREADIMAGE               clEnqueueReadImage              = NULL;
PFNCLENQUEUEWRITEIMAGE              clEnqueueWriteImage             = NULL;
PFNCLENQUEUECOPYIMAGE               clEnqueueCopyImage              = NULL;
PFNCLENQUEUECOPYIMAGETOBUFFER       clEnqueueCopyImageToBuffer      = NULL;
PFNCLENQUEUECOPYBUFFERTOIMAGE       clEnqueueCopyBufferToImage      = NULL;
PFNCLENQUEUEMAPBUFFER               clEnqueueMapBuffer              = NULL;
PFNCLENQUEUEMAPIMAGE                clEnqueueMapImage               = NULL;
PFNCLENQUEUEUNMAPMEMOBJECT          clEnqueueUnmapMemObject         = NULL;
PFNCLENQUEUENDRANGEKERNEL           clEnqueueNDRangeKernel          = NULL;
PFNCLENQUEUETASK                    clEnqueueTask                   = NULL;
PFNCLENQUEUENATIVEKERNEL            clEnqueueNativeKernel           = NULL;
PFNCLENQUEUEMARKER                  clEnqueueMarker                 = NULL;
PFNCLENQUEUEWAITFOREVENTS           clEnqueueWaitForEvents          = NULL;
PFNCLENQUEUEBARRIER                 clEnqueueBarrier                = NULL;
PFNCLGETEXTENSIONFUNCTIONADDRESS    clGetExtensionFunctionAddress   = NULL;

void FreeCL(void)
    if ($clDllModule != NULL)
        FreeLibrary($clDllModule);
        $clDllModule = NULL;

int LoadCL(char* path)
    int error = 0;
    //  Check if already initialized
    if ($clDllModule != NULL)
        return 1;

    //  Load library
    $clDllModule = LoadLibraryA(path);

    //  Check for errors
    if ($clDllModule == NULL)
        return 0;

    //  Set unloading
    error = atexit(FreeCL);

    if (error)
        //  Failure queing atexit, shutdown with error
        FreeLibrary($clDllModule);
        $clDllModule = NULL;

        return 0;

    //  Determine function entry-points
    $clGetPlatformIDs                = (PFNCLGETPLATFORMIDS              )GetProcAddress($clDllModule, "clGetPlatformIDs");
    $clGetPlatformInfo               = (PFNCLGETPLATFORMINFO             )GetProcAddress($clDllModule, "clGetPlatformInfo");
    $clGetDeviceIDs                  = (PFNCLGETDEVICEIDS                )GetProcAddress($clDllModule, "clGetDeviceIDs");
    $clGetDeviceInfo                 = (PFNCLGETDEVICEINFO               )GetProcAddress($clDllModule, "clGetDeviceInfo");
    $clCreateContext                 = (PFNCLCREATECONTEXT               )GetProcAddress($clDllModule, "clCreateContext");
    $clCreateContextFromType         = (PFNCLCREATECONTEXTFROMTYPE       )GetProcAddress($clDllModule, "clCreateContextFromType");
    $clRetainContext                 = (PFNCLRETAINCONTEXT               )GetProcAddress($clDllModule, "clRetainContext");
    $clReleaseContext                = (PFNCLRELEASECONTEXT              )GetProcAddress($clDllModule, "clReleaseContext");
    $clGetContextInfo                = (PFNCLGETCONTEXTINFO              )GetProcAddress($clDllModule, "clGetContextInfo");
    $clCreateCommandQueue            = (PFNCLCREATECOMMANDQUEUE          )GetProcAddress($clDllModule, "clCreateCommandQueue");
    $clRetainCommandQueue            = (PFNCLRETAINCOMMANDQUEUE          )GetProcAddress($clDllModule, "clRetainCommandQueue");
    $clReleaseCommandQueue           = (PFNCLRELEASECOMMANDQUEUE         )GetProcAddress($clDllModule, "clReleaseCommandQueue");
    $clGetCommandQueueInfo           = (PFNCLGETCOMMANDQUEUEINFO         )GetProcAddress($clDllModule, "clGetCommandQueueInfo");
    $clSetCommandQueueProperty       = (PFNCLSETCOMMANDQUEUEPROPERTY     )GetProcAddress($clDllModule, "clSetCommandQueueProperty");
    $clCreateBuffer                  = (PFNCLCREATEBUFFER                )GetProcAddress($clDllModule, "clCreateBuffer");
    $clCreateImage2D                 = (PFNCLCREATEIMAGE2D               )GetProcAddress($clDllModule, "clCreateImage2D");
    $clCreateImage3D                 = (PFNCLCREATEIMAGE3D               )GetProcAddress($clDllModule, "clCreateImage3D");
    $clRetainMemObject               = (PFNCLRETAINMEMOBJECT             )GetProcAddress($clDllModule, "clRetainMemObject");
    $clReleaseMemObject              = (PFNCLRELEASEMEMOBJECT            )GetProcAddress($clDllModule, "clReleaseMemObject");
    $clGetSupportedImageFormats      = (PFNCLGETSUPPORTEDIMAGEFORMATS    )GetProcAddress($clDllModule, "clGetSupportedImageFormats");
    $clGetMemObjectInfo              = (PFNCLGETMEMOBJECTINFO            )GetProcAddress($clDllModule, "clGetMemObjectInfo");
    $clGetImageInfo                  = (PFNCLGETIMAGEINFO                )GetProcAddress($clDllModule, "clGetImageInfo");
    $clCreateSampler                 = (PFNCLCREATESAMPLER               )GetProcAddress($clDllModule, "clCreateSampler");
    $clRetainSampler                 = (PFNCLRETAINSAMPLER               )GetProcAddress($clDllModule, "clRetainSampler");
    $clReleaseSampler                = (PFNCLRELEASESAMPLER              )GetProcAddress($clDllModule, "clReleaseSampler");
    $clGetSamplerInfo                = (PFNCLGETSAMPLERINFO              )GetProcAddress($clDllModule, "clGetSamplerInfo");
    $clCreateProgramWithSource       = (PFNCLCREATEPROGRAMWITHSOURCE     )GetProcAddress($clDllModule, "clCreateProgramWithSource");
    $clCreateProgramWithBinary       = (PFNCLCREATEPROGRAMWITHBINARY     )GetProcAddress($clDllModule, "clCreateProgramWithBinary");
    $clRetainProgram                 = (PFNCLRETAINPROGRAM               )GetProcAddress($clDllModule, "clRetainProgram");
    $clReleaseProgram                = (PFNCLRELEASEPROGRAM              )GetProcAddress($clDllModule, "clReleaseProgram");
    $clBuildProgram                  = (PFNCLBUILDPROGRAM                )GetProcAddress($clDllModule, "clBuildProgram");
    $clUnloadCompiler                = (PFNCLUNLOADCOMPILER              )GetProcAddress($clDllModule, "clUnloadCompiler");
    $clGetProgramInfo                = (PFNCLGETPROGRAMINFO              )GetProcAddress($clDllModule, "clGetProgramInfo");
    $clGetProgramBuildInfo           = (PFNCLGETPROGRAMBUILDINFO         )GetProcAddress($clDllModule, "clGetProgramBuildInfo");
    $clCreateKernel                  = (PFNCLCREATEKERNEL                )GetProcAddress($clDllModule, "clCreateKernel");
    $clCreateKernelsInProgram        = (PFNCLCREATEKERNELSINPROGRAM      )GetProcAddress($clDllModule, "clCreateKernelsInProgram");
    $clRetainKernel                  = (PFNCLRETAINKERNEL                )GetProcAddress($clDllModule, "clRetainKernel");
    $clReleaseKernel                 = (PFNCLRELEASEKERNEL               )GetProcAddress($clDllModule, "clReleaseKernel");
    $clSetKernelArg                  = (PFNCLSETKERNELARG                )GetProcAddress($clDllModule, "clSetKernelArg");
    $clGetKernelInfo                 = (PFNCLGETKERNELINFO               )GetProcAddress($clDllModule, "clGetKernelInfo");
    $clGetKernelWorkGroupInfo        = (PFNCLGETKERNELWORKGROUPINFO      )GetProcAddress($clDllModule, "clGetKernelWorkGroupInfo");
    $clWaitForEvents                 = (PFNCLWAITFOREVENTS               )GetProcAddress($clDllModule, "clWaitForEvents");
    $clGetEventInfo                  = (PFNCLGETEVENTINFO                )GetProcAddress($clDllModule, "clGetEventInfo");
    $clRetainEvent                   = (PFNCLRETAINEVENT                 )GetProcAddress($clDllModule, "clRetainEvent");
    $clReleaseEvent                  = (PFNCLRELEASEEVENT                )GetProcAddress($clDllModule, "clReleaseEvent");
    $clGetEventProfilingInfo         = (PFNCLGETEVENTPROFILINGINFO       )GetProcAddress($clDllModule, "clGetEventProfilingInfo");
    $clFlush                         = (PFNCLFLUSH                       )GetProcAddress($clDllModule, "clFlush");
    $clFinish                        = (PFNCLFINISH                      )GetProcAddress($clDllModule, "clFinish");
    $clEnqueueReadBuffer             = (PFNCLENQUEUEREADBUFFER           )GetProcAddress($clDllModule, "clEnqueueReadBuffer");
    $clEnqueueWriteBuffer            = (PFNCLENQUEUEWRITEBUFFER          )GetProcAddress($clDllModule, "clEnqueueWriteBuffer");
    $clEnqueueCopyBuffer             = (PFNCLENQUEUECOPYBUFFER           )GetProcAddress($clDllModule, "clEnqueueCopyBuffer");
    $clEnqueueReadImage              = (PFNCLENQUEUEREADIMAGE            )GetProcAddress($clDllModule, "clEnqueueReadImage");
    $clEnqueueWriteImage             = (PFNCLENQUEUEWRITEIMAGE           )GetProcAddress($clDllModule, "clEnqueueWriteImage");
    $clEnqueueCopyImage              = (PFNCLENQUEUECOPYIMAGE            )GetProcAddress($clDllModule, "clEnqueueCopyImage");
    $clEnqueueCopyImageToBuffer      = (PFNCLENQUEUECOPYIMAGETOBUFFER    )GetProcAddress($clDllModule, "clEnqueueCopyImageToBuffer");
    $clEnqueueCopyBufferToImage      = (PFNCLENQUEUECOPYBUFFERTOIMAGE    )GetProcAddress($clDllModule, "clEnqueueCopyBufferToImage");
    $clEnqueueMapBuffer              = (PFNCLENQUEUEMAPBUFFER            )GetProcAddress($clDllModule, "clEnqueueMapBuffer");
    $clEnqueueMapImage               = (PFNCLENQUEUEMAPIMAGE             )GetProcAddress($clDllModule, "clEnqueueMapImage");
    $clEnqueueUnmapMemObject         = (PFNCLENQUEUEUNMAPMEMOBJECT       )GetProcAddress($clDllModule, "clEnqueueUnmapMemObject");
    $clEnqueueNDRangeKernel          = (PFNCLENQUEUENDRANGEKERNEL        )GetProcAddress($clDllModule, "clEnqueueNDRangeKernel");
    $clEnqueueTask                   = (PFNCLENQUEUETASK                 )GetProcAddress($clDllModule, "clEnqueueTask");
    $clEnqueueNativeKernel           = (PFNCLENQUEUENATIVEKERNEL         )GetProcAddress($clDllModule, "clEnqueueNativeKernel");
    $clEnqueueMarker                 = (PFNCLENQUEUEMARKER               )GetProcAddress($clDllModule, "clEnqueueMarker");
    $clEnqueueWaitForEvents          = (PFNCLENQUEUEWAITFOREVENTS        )GetProcAddress($clDllModule, "clEnqueueWaitForEvents");
    $clEnqueueBarrier                = (PFNCLENQUEUEBARRIER              )GetProcAddress($clDllModule, "clEnqueueBarrier");
    $clGetExtensionFunctionAddress   = (PFNCLGETEXTENSIONFUNCTIONADDRESS )GetProcAddress($clDllModule, "clGetExtensionFunctionAddress");

    return 1;

//! \param error CL error code
//! \return a string representation of the error code
char* clErrorString(cl_int error)
    const char*[] strings = char *[] const[
        // Error Codes
          "CL_SUCCESS"                                  //   0
        , "CL_DEVICE_NOT_FOUND"                         //  -1
        , "CL_DEVICE_NOT_AVAILABLE"                     //  -2
        , "CL_COMPILER_NOT_AVAILABLE"                   //  -3
        , "CL_MEM_OBJECT_ALLOCATION_FAILURE"            //  -4
        , "CL_OUT_OF_RESOURCES"                         //  -5
        , "CL_OUT_OF_HOST_MEMORY"                       //  -6
        , "CL_PROFILING_INFO_NOT_AVAILABLE"             //  -7
        , "CL_MEM_COPY_OVERLAP"                         //  -8
        , "CL_IMAGE_FORMAT_MISMATCH"                    //  -9
        , "CL_IMAGE_FORMAT_NOT_SUPPORTED"               //  -10
        , "CL_BUILD_PROGRAM_FAILURE"                    //  -11
        , "CL_MAP_FAILURE"                              //  -12

        , ""    //  -13
        , ""    //  -14
        , ""    //  -15
        , ""    //  -16
        , ""    //  -17
        , ""    //  -18
        , ""    //  -19

        , ""    //  -20
        , ""    //  -21
        , ""    //  -22
        , ""    //  -23
        , ""    //  -24
        , ""    //  -25
        , ""    //  -26
        , ""    //  -27
        , ""    //  -28
        , ""    //  -29

        , "CL_INVALID_VALUE"                            //  -30
        , "CL_INVALID_DEVICE_TYPE"                      //  -31
        , "CL_INVALID_PLATFORM"                         //  -32
        , "CL_INVALID_DEVICE"                           //  -33
        , "CL_INVALID_CONTEXT"                          //  -34
        , "CL_INVALID_QUEUE_PROPERTIES"                 //  -35
        , "CL_INVALID_COMMAND_QUEUE"                    //  -36
        , "CL_INVALID_HOST_PTR"                         //  -37
        , "CL_INVALID_MEM_OBJECT"                       //  -38
        , "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR"          //  -39
        , "CL_INVALID_IMAGE_SIZE"                       //  -40
        , "CL_INVALID_SAMPLER"                          //  -41
        , "CL_INVALID_BINARY"                           //  -42
        , "CL_INVALID_BUILD_OPTIONS"                    //  -43
        , "CL_INVALID_PROGRAM"                          //  -44
        , "CL_INVALID_PROGRAM_EXECUTABLE"               //  -45
        , "CL_INVALID_KERNEL_NAME"                      //  -46
        , "CL_INVALID_KERNEL_DEFINITION"                //  -47
        , "CL_INVALID_KERNEL"                           //  -48
        , "CL_INVALID_ARG_INDEX"                        //  -49
        , "CL_INVALID_ARG_VALUE"                        //  -50
        , "CL_INVALID_ARG_SIZE"                         //  -51
        , "CL_INVALID_KERNEL_ARGS"                      //  -52
        , "CL_INVALID_WORK_DIMENSION"                   //  -53
        , "CL_INVALID_WORK_GROUP_SIZE"                  //  -54
        , "CL_INVALID_WORK_ITEM_SIZE"                   //  -55
        , "CL_INVALID_GLOBAL_OFFSET"                    //  -56
        , "CL_INVALID_EVENT_WAIT_LIST"                  //  -57
        , "CL_INVALID_EVENT"                            //  -58
        , "CL_INVALID_OPERATION"                        //  -59
        , "CL_INVALID_GL_OBJECT"                        //  -60
        , "CL_INVALID_BUFFER_SIZE"                      //  -61
        , "CL_INVALID_MIP_LEVEL"                        //  -62
        , "CL_INVALID_GLOBAL_WORK_SIZE"                 //  -63
	];
    return strings[min(max(-error,0),63)];
