uses spap, cl

// texture access intrinsics

__intrinsic__ int UseTextureRead(cl_mem texId, cl_sampler sampler)
	/meta.state("clbe-useTextureRead") = meta.curfunc
__intrinsic__ int UseTextureWrite(cl_mem texId)
	/meta.state("clbe-useTextureWrite") = meta.curfunc
__device__ __intrinsic__ int Image2dFromHandle(int texId)
	/meta.state("clbe-image2dFromHandle") = meta.curfunc
__device__ __intrinsic__ int Image3dFromHandle(int texId)
	/meta.state("clbe-image3dFromHandle") = meta.curfunc
__device__ __intrinsic__ int SamplerFromHandle(int sampler)
	/meta.state("clbe-samplerFromHandle") = meta.curfunc

// OpenCL Build-in types
// For internal use only
class image2d_t
	int dummy;
	/{meta.curstruct.cbetags|=meta.cstruct.CBE_CUDASYS}
	/meta.state("clbe-type-image2d_t") = (void*)meta.curstruct.gettype()
class image3d_t
	int dummy;
	/{meta.curstruct.cbetags|=meta.cstruct.CBE_CUDASYS}
	/meta.state("clbe-type-image3d_t") = (void*)meta.curstruct.gettype()
class sampler_t
	int dummy;
	/{meta.curstruct.cbetags|=meta.cstruct.CBE_CUDASYS}
	/meta.state("clbe-type-sampler_t") = (void*)meta.curstruct.gettype()

import "clbe.dll"
	void prepare()
/
	prepare()
	meta.enableTarget("OpenCL",1)
	meta.enableTarget("CPU",0)

static detail
using detail

with detail

	void transhared(iptr s)
		using meta
		v=(cvar*)s
		feedexpr(&call("operator*",List[&makeerv(v)]))
#def __shared__:
	using meta
	t=atype()
	id=wantid()
	if got(int'['):
		t=typemodifiers(t)
	v=makevar(makeptrtype(t))
	emitcode_clear()
	emitcode_expr(castto(&makeerv(v),t0_iptr))
	write-emitcodef(;"=$$spap.detail.clbe_allocsm(",_sizeof(t),",",alignof(t),");spapTargetOption('forceTarget','OpenCL');")
	emitcode_flush()
	declare_metaid(id,detail.transhared,(iptr)v)

cl_context clGetContext()
	f = (cl_context()() spapGetGpuApi("clGetContext"));
	if (f)
		return f()
	else
		return NULL;
cl_command_queue clGetQueue() = (cl_command_queue()() spapGetGpuApi("clGetQueue"))()

// Texture class
//   Possible template values:
//   TDim = {2, 3}
//   channel = {1,2,3,4, CL_R, CL_A, CL_RG, CL_RGB, CL_RGBA}
//   dataType = {CL_HALF_FLOAT, CL_FLOAT, CL_SNORM_INT8/16, CL_UNORM_INT8/16, CL_UNSIGNED_INT8/16/32, CL_SIGNED_INT8/16/32}
class CLTextureBase{
	cl_mem texId;
	cl_sampler sampler;
	cl_image_format format;
}

class CLTexture(const TDim, const channel, const dataType) : CLTextureBase{
	If (u32 dataType == CL_SNORM_INT8 || u32 dataType == CL_SNORM_INT16 || u32 dataType == CL_UNORM_INT8 || u32 dataType == CL_UNORM_INT16
		|| u32 dataType == CL_UNORM_SHORT_565 || u32 dataType == CL_UNORM_SHORT_555 || u32 dataType == CL_UNORM_INT_101010 ||
		u32 dataType == CL_HALF_FLOAT || u32 dataType == CL_FLOAT)
		typedef float BaseT;
	Elif (u32 dataType == CL_UNSIGNED_INT8 || u32 dataType == CL_UNSIGNED_INT16 || u32 dataType == CL_UNSIGNED_INT32)
		typedef u32 BaseT;
	Else
		typedef int BaseT;
	inline novirtual u32 TranslateChannel()
		If (channel == 1)
			return CL_R;
		Elif (channel == 2)
			return CL_RG;
		Elif (channel == 3)
			return CL_RGB;
		Elif (channel == 4)
			return CL_RGBA;
		return (u32)channel;
	novirtual void SetDefaultSampler()
		if (.sampler)
			cl.clReleaseSampler(.sampler);
		.sampler = cl.clCreateSampler(clGetContext(), CL_TRUE, CL_ADDRESS_CLAMP_TO_EDGE, CL_FILTER_NEAREST, NULL)
	novirtual void SetSampler(cl_bool normalized_coords, cl_addressing_mode addrMode, cl_filter_mode filterMode)
		if (.sampler)
			cl.clReleaseSampler(.sampler);
		.sampler = cl.clCreateSampler(clGetContext(), normalized_coords, addrMode, filterMode, NULL);
If (TDim == 2) {
	__init__(int w, int h)
		.sampler = NULL;
		.format.image_channel_order = .TranslateChannel();
		.format.image_channel_data_type = (u32)dataType;
		ctx = clGetContext();
		.texId = cl.clCreateImage2D(clGetContext(), CL_MEM_READ_WRITE, &.format, (size_t)w, (size_t)h, (size_t)0, NULL, NULL);
		.SetDefaultSampler();
	novirtual void Resize(int w, int h)
		if (.texId)
			cl.clReleaseMemObject(.texId);
		.texId = cl.clCreateImage2D(clGetContext(), CL_MEM_READ_WRITE, &.format, (size_t)w, (size_t)h, (size_t)0, NULL, NULL);
}
Elif (TDim == 3) {
	__init__(int w, int h, int d)
		.sampler = NULL;
		.format.image_channel_order = .TranslateChannel();
		.format.image_channel_data_type = (u32)dataType;
		.texId = cl.clCreateImage3D(clGetContext(), CL_MEM_READ_WRITE, &.format, (size_t)w, (size_t)h, (size_t)d, (size_t)0, (size_t) 0, NULL, NULL);
		.SetDefaultSampler();
	novirtual void Resize(int w, int h)
		if (.texId)
			cl.clReleaseMemObject(.texId);
		.texId = cl.clCreateImage3D(clGetContext(), CL_MEM_READ_WRITE, &.format, (size_t)w, (size_t)h, (size_t)d, (size_t)0, (size_t) 0, NULL, NULL);
}
	__done__()
		cl.clReleaseMemObject(.texId);
		cl.clReleaseSampler(.sampler);

	If (meta.typeid(BaseT) == meta.typeid(float))
		typedef float4 TexelType
	Elif (meta.typeid(BaseT) == meta.typeid(int))
		typedef int4 TexelType
	Else
		typedef uint4 TexelType
	If (TDim == 2)
		typedef int2 CoordTypeI
		typedef float2 CoordTypeF
	Else
		typedef int4 CoordTypeI
		typedef float4 CoordTypeF

	__operation__ TexelType Fetch(CoordTypeI coord)
		case "OpenCL"
			int texHandle = UseTextureRead(.texId, .sampler);
			spawn ...
				...
				__asm__("@return",meta.typeid(float4),"@returnto",(TexelType*)result,
					"=read_imagef(",Image2dFromHandle(texHandle),",", SamplerFromHandle(texHandle), ",", coord , ")");
				...
	__operation__ TexelType Fetch(CoordTypeF coord)
		case "OpenCL"
			int texHandle = UseTextureRead(.texId, .sampler);
			spawn ...
				...
				__asm__("@return",meta.typeid(float4),"@returnto",(TexelType*)result,
					"=read_imagef(",Image2dFromHandle(texHandle),",", SamplerFromHandle(texHandle), ",", coord , ")");
				...
	__operation__ void Write(CoordTypeI coord, TexelType color)
		case "OpenCL"
			int texHandle = UseTextureWrite(.texId);
			spawn ...
				...
				__asm__("write_imagef(",Image2dFromHandle(texHandle),",", coord, ",", color, ");", "@prelive");
				...
	void ReadImage(void * buf, int3 origin, int3 size, int rowPitch, int slicePitch)
		size_t[3] _origin, _size;
		For i = 0:2
			_origin[i] = (size_t)origin[i];
			_size[i] = (size_t)size[i];
		cl.clEnqueueReadImage(clGetQueue(), .texId, CL_TRUE, _origin, _size, rowPitch, slicePitch, buf, u32 0, NULL, NULL);
	void WriteImage(void * buf, int3 origin, int3 size, int rowPitch, int slicePitch)
		size_t[3] _origin, _size;
		For i = 0:2
			_origin[i] = (size_t)origin[i];
			_size[i] = (size_t)size[i];
		cl.clEnqueueWriteImage(clGetQueue(), .texId, CL_TRUE, _origin, _size, rowPitch, slicePitch, buf, u32 0, NULL, NULL);
	void CopyFrom(CLTextureBase src, int3 srcOrigin, int3 destOrigin, int3 size)
		size_t[3] _originSrc, _originDest, _size;
		For i = 0:2
			_originSrc[i] = (size_t)srcOrigin[i];
			_originDest[i] = (size_t)destOrigin[i];
			_size[i] = (size_t)size[i];
		queue = clGetQueue();
		cl.clEnqueueCopyImage(queue, src.texId, .texId, _originSrc, _originDest, _size, u32 0, NULL, NULL);
		cl.clEnqueueBarrier(queue);
}

%__init__()
	cl.LoadCL("OpenCL.dll");