/*
 * Copyright 1993-2009 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

/*
* cufft.h
* Public header file for the NVIDIA Cuda FFT library (CUFFT)
*/

//contributed by Zhong Ren

uses cuda

const
	CUFFT_SUCCESS		=0x00000000
	CUFFT_INVALID_PLAN	=0x00000001
	CUFFT_ALLOC_FAILED  =0x00000002
	CUFFT_INVALID_TYPE	=0x00000003
	CUFFT_INVALID_VALUE =0x00000004
	CUFFT_INTERNAL_ERROR=0x00000005
	CUFFT_EXEC_FAILED	=0x00000006
	CUFFT_SETUP_FAILED	=0x00000007
	CUFFT_INVALID_SIZE	=0x00000008

// CUFFT defines and supports the following data types

// CUFFT transform directions
const CUFFT_FORWARD=-1 // Forward FFT
const CUFFT_INVERSE=1 // Inverse FFT

// CUFFT supports the following transform types
const
  CUFFT_R2C = 0x2a     // Real to Complex (interleaved)
  CUFFT_C2R = 0x2c     // Complex (interleaved) to Real
  CUFFT_C2C = 0x29     // Complex to Complex, interleaved
  CUFFT_D2Z = 0x6a     // Double to Double-Complex
  CUFFT_Z2D = 0x6c     // Double-Complex to Double
  CUFFT_Z2Z = 0x69     // Double-Complex to Double-Complex

import "cufft.dll"{

int stdcall cufftPlan1d(unsigned int *plan,
                                 int nx,
                                 int type,
                                 int batch);

int stdcall cufftPlan2d(unsigned int *plan,
                                 int nx, int ny,
                                 int type);

int stdcall cufftPlan3d(unsigned int *plan,
                                 int nx, int ny, int nz,
                                 int type);

int stdcall cufftDestroy(unsigned int plan);

int stdcall cufftExecC2C(unsigned int plan,
                                  float *idata,
                                  float *odata,
                                  int direction);

int stdcall cufftExecR2C(unsigned int plan,
                                  float *idata,
                                  float *odata);

int stdcall cufftExecC2R(unsigned int plan,
                                  float *idata,
                                  float *odata);


int stdcall cufftExecZ2Z(unsigned int plan,
                                  double *idata,
                                  double *odata,
                                  int direction);

int stdcall cufftExecD2Z(unsigned int plan,
                                  double *idata,
                                  double *odata);

int stdcall cufftExecZ2D(unsigned int plan,
                                  double *idata,
                                  double *odata);

//int stdcall cufftSetStream(unsigned int p, cudaStream_t stream);
}


